"""StatusMessage adapter tests."""

import unittest


class TestAdapter(unittest.TestCase):
    def test_directives(self):
        """
        Test status messages

        First some boilerplate.

          >>> from zope.component.testing import setUp
          >>> setUp()

          >>> import Products.Five
          >>> import Products.statusmessages

          >>> from Zope2.App import zcml
          >>> zcml.load_config('meta.zcml', Products.Five)
          >>> zcml.load_config('configure.zcml', Products.statusmessages)

        Now lets make sure we can actually adapt the request.

          >>> from Products.statusmessages.interfaces import IStatusMessage
          >>> status = IStatusMessage(self.app.REQUEST)
          >>> IStatusMessage.providedBy(status)
          True

        We also need the request to be annotatable:

          >>> from zope.interface import directlyProvides
          >>> from zope.annotation.interfaces import IAttributeAnnotatable
          >>> directlyProvides(self.app.REQUEST, IAttributeAnnotatable)

        The dummy request we have is a bit limited, so we need a simple method
        to fake a real request/response for the cookie handling. Basically it
        puts all entries from response.cookies into REQUEST.cookies but shifts
        the real values into the right place as browsers would do it.

          >>> def fakePublish(request):
          ...     cookies = request.response.cookies.copy()
          ...     new_cookies = {}
          ...     for key in cookies.keys():
          ...         new_cookies[key] = cookies[key]['value']
          ...     request.cookies = new_cookies
          ...     request.response.cookies = {}

          >>> request = self.app.REQUEST
          >>> status = IStatusMessage(request)

        Make sure there's no stored message.

          >>> len(status.show())
          0

        Add one message

          >>> status.add(u'täst', type=u'info')

        Now check the results

          >>> messages = status.show()
          >>> len(messages)
          1

          >>> messages[0].message
          u'täst'

          >>> messages[0].type
          u'info'

        Make sure messages are removed

          >>> len(status.show())
          0

        Since we accessed the message prior to publishing the page, we must
        ensure that the messages have been removed from the cookies

          >>> fakePublish(request)
          >>> len(status.show())
          0

        Now we repeat the test, only this time we publish the page prior to
        retrieving the messages

        Add one message

          >>> status.add(u'täst', type=u'info')

        Publish the request

          >>> fakePublish(request)

        Now check the results

          >>> messages = status.show()
          >>> len(messages)
          1

          >>> messages[0].message
          u'täst'

          >>> messages[0].type
          u'info'

        Make sure messages are removed

          >>> len(status.show())
          0

        Add two messages (without publishing)

          >>> status.add(u'täst', type=u'info')
          >>> status.add(u'täst1', u'warn')

        And check the results again

          >>> messages = status.show()
          >>> len(messages)
          2

          >>> test = messages[1]

          >>> test.message
          u'täst1'

          >>> test.type
          u'warn'

        Make sure messages are removed again

          >>> len(status.show())
          0

        Add two messages (with publishing)

          >>> status.add(u'täst', type=u'info')
          >>> fakePublish(request)
          >>> status.add(u'test1', u'warn')

        And check the results again

          >>> fakePublish(request)
          >>> messages = status.show()
          >>> len(messages)
          2

          >>> test = messages[1]

          >>> test.message
          u'test1'

          >>> test.type
          u'warn'

        Make sure messages are removed again

          >>> len(status.show())
          0

        Add two identical messages

          >>> status.add(u'täst', type=u'info')
          >>> status.add(u'täst', type=u'info')

        And check the results again

          >>> fakePublish(request)
          >>> messages = status.show()
          >>> len(messages)
          1

          >>> test = messages[0]

          >>> test.message
          u'täst'

          >>> test.type
          u'info'

        Make sure messages are removed again

          >>> len(status.show())
          0

        Test incredibly long messages:

          >>> status.add(u'm' * 0x400, type=u't' * 0x20)

          And check the results again

          >>> fakePublish(request)
          >>> messages = status.show()
          >>> len(messages)
          1

          >>> test = messages[0]

          >>> test.message == u'm' * 0x3FF
          True

          >>> test.type == u't' * 0x1F
          True

        Messages are stored as base64-ed cookie values, so we must make sure we
        create proper header values; all ascii characters, and no newlines:

          >>> status.add(u'täst' * 40, type=u'info')
          >>> cookies = [c['value'] for c in request.response.cookies.values()]
          >>> cookies = ''.join(cookies)
          >>> cookies == str(cookies).encode('ASCII')
          True
          >>> '\\n' in cookies
          False

          >>> from zope.component.testing import tearDown
          >>> tearDown()
        """

    def test_301(self):
        """
        Test status messages for 301/302/304 request

        First some boilerplate.

          >>> from zope.component.testing import setUp
          >>> setUp()

          >>> import Products.Five
          >>> import Products.statusmessages

          >>> from Zope2.App import zcml
          >>> zcml.load_config('meta.zcml', Products.Five)
          >>> zcml.load_config('configure.zcml', Products.statusmessages)

          >>> from zope.interface import directlyProvides
          >>> from zope.annotation.interfaces import IAttributeAnnotatable
          >>> directlyProvides(self.app.REQUEST, IAttributeAnnotatable)

          >>> from Products.statusmessages.interfaces import IStatusMessage

          >>> def fakePublish(request, status=200):
          ...     cookies = request.response.cookies.copy()
          ...     new_cookies = {}
          ...     for key in cookies.keys():
          ...         new_cookies[key] = cookies[key]['value']
          ...     request.cookies = new_cookies
          ...     request.response.cookies = {}
          ...     request.response.setStatus(status)

          >>> request = self.app.REQUEST
          >>> status = IStatusMessage(request)

        Make sure there's no stored message.

          >>> len(status.show())
          0

        Add one message

          >>> status.add(u'täst', type=u'info')

        Publish a redirect response that also happened to call show().
        This could happen if the redirect (unnecessarily)
        rendered a template showing the status message, for example.

          >>> fakePublish(request, 302)
          >>> messages = status.show()
          >>> len(messages)
          1

          >>> messages[0].message
          u'täst'

          >>> messages[0].type
          u'info'

        Make sure messages are not removed - we really want them to show the
        next time around, when the redirect has completed.

          >>> len(status.show())
          1

        Let's now fake redirection. The message should still be there,
        but will then be expired.

          >>> fakePublish(request, 200)
          >>> messages = status.show()
          >>> len(messages)
          1

          >>> messages[0].message
          u'täst'

          >>> messages[0].type
          u'info'

        The message should now be gone.

          >>> len(status.show())
          0

          >>> from zope.component.testing import tearDown
          >>> tearDown()
        """



import unittest


class TestEncoding(unittest.TestCase):
    def test_encoding_msg_with_type(self):
        """Test message encoding:"""
        from Products.statusmessages.message import decode
        from Products.statusmessages.message import Message

        m = Message("späm", "eggs")
        self.assertEqual(
            m.encode(),
            b"\x00\xa4sp\xc3\xa4meggs",
        )
        self.assertEqual(decode(m.encode())[0], m)

    def test_encoding_msg_without_type(self):
        from Products.statusmessages.message import decode
        from Products.statusmessages.message import Message

        m = Message("späm")
        self.assertEqual(
            m,
            Message("späm"),
        )
        self.assertEqual(m.encode(), b"\x00\xa0sp\xc3\xa4m")
        self.assertEqual(decode(m.encode())[0], m)

    def test_decoding(self):
        """Test message decoding:"""
        from Products.statusmessages.message import decode

        # Craft a wrong value:
        m, rem = decode(b"\x01\x84spameggs")
        self.assertEqual(
            m.message,
            "spameggs",
        )
        self.assertEqual(
            m.type,
            "",
        )
        self.assertEqual(rem, b"")

        # Craft another wrong value:
        m, rem = decode(b"\x00\x24spameggs")
        self.assertEqual(
            m.message,
            "s",
        )
        self.assertEqual(
            m.type,
            "pame",
        )
        self.assertEqual(rem, b"ggs")

        # And another wrong value:
        m, rem = decode(b"\x00spameggs")
        self.assertEqual(
            m.message,
            "pam",
        )
        self.assertEqual(
            m.type,
            "eggs",
        )
        self.assertEqual(rem, b"")

        # And yet another wrong value:
        m, rem = decode("")

        self.assertIs(m, None)
        self.assertEqual(rem, b"")



from zope.interface import Attribute
from zope.interface import Interface


class IMessage(Interface):
    """A single status message."""

    message = Attribute("The text of this message. Usually a Message object.")

    type = Attribute("The type of this message.")


class IStatusMessage(Interface):
    """An adapter for the BrowserRequest to handle status messages."""

    def addStatusMessage(text, type="info"):
        """Add a status message."""

    def add(text, type="info"):
        """Add a status message."""

    def showStatusMessages():
        """Removes all status messages and returns them for display."""

    def show():
        """Removes all status messages and returns them for display."""



from Products.statusmessages import STATUSMESSAGEKEY
from Products.statusmessages.interfaces import IStatusMessage
from Products.statusmessages.message import decode
from Products.statusmessages.message import Message
from zope.annotation.interfaces import IAnnotations
from zope.i18n import translate
from zope.interface import implementer

import binascii
import logging


logger = logging.getLogger("statusmessages")


@implementer(IStatusMessage)
class StatusMessage:
    """Adapter for the BrowserRequest to handle status messages.

    Let's make sure that this implementation actually fulfills the
    'IStatusMessage' API::

        >>> from zope.interface.verify import verifyClass
        >>> verifyClass(IStatusMessage, StatusMessage)
        True
    """

    def __init__(self, context):
        self.context = context  # the context must be the request

    def add(self, text, type="info"):
        """Add a status message."""
        context = self.context
        text = translate(text, context=context)
        annotations = IAnnotations(context)

        old = annotations.get(
            STATUSMESSAGEKEY,
            context.cookies.get(STATUSMESSAGEKEY),
        )
        value = _encodeCookieValue(text, type, old=old)
        context.response.setCookie(STATUSMESSAGEKEY, value, path="/")
        annotations[STATUSMESSAGEKEY] = value

    def show(self):
        """Removes all status messages and returns them for display."""
        context = self.context
        annotations = IAnnotations(context)
        value = annotations.get(
            STATUSMESSAGEKEY,
            context.cookies.get(STATUSMESSAGEKEY),
        )
        if value is None:
            return []
        value = _decodeCookieValue(value)

        # clear the existing cookie entries, except on responses that don't
        # actually render in the browser (really, these shouldn't render
        # anything so we shouldn't get to this message, but some templates
        # are sloppy).
        if self.context.response.getStatus() not in (301, 302, 304):
            context.cookies[STATUSMESSAGEKEY] = None
            context.response.expireCookie(STATUSMESSAGEKEY, path="/")
            annotations[STATUSMESSAGEKEY] = None

        return value

    # BBB
    addStatusMessage = add
    showStatusMessages = show


def _encodeCookieValue(text, type, old=None):
    """Encodes text and type to a list of Messages. If there is already some old
    existing list, add the new Message at the end but don't add duplicate
    messages.
    """
    results = []
    message = Message(text, type=type)

    if old is not None:
        results = _decodeCookieValue(old)
    if message not in results:
        results.append(message)

    messages = b"".join([r.encode() for r in results])
    bin_value = binascii.b2a_base64(messages).rstrip()
    # remove the stupid b that will lead to values like "b'AYR...'"
    return bin_value.decode("utf-8")


def _decodeCookieValue(string):
    """Decode a cookie value to a list of Messages."""
    results = []
    # Return nothing if the cookie is marked as deleted
    if string == "deleted":
        return results
    # Try to decode the cookie value
    try:
        value = binascii.a2b_base64(string)
        while len(value) > 1:  # at least 2 bytes of data
            message, value = decode(value)
            if message is not None:
                results.append(message)
    except (binascii.Error, UnicodeEncodeError):
        logger.exception("Unexpected value in statusmessages cookie")
        return []

    return results



from doctest import DocTestSuite
from unittest import TestSuite


test_list = (
    DocTestSuite("Products.statusmessages.adapter"),
    DocTestSuite("Products.statusmessages.message"),
)


def test_suite():
    return TestSuite(test_list)



from Products.statusmessages.interfaces import IMessage
from zope.interface import implementer

import struct


def _utf8(value):
    if isinstance(value, str):
        return value.encode("utf-8")
    elif isinstance(value, bytes):
        return value
    return b""


def _unicode(value):
    return str(value, "utf-8", "ignore")


@implementer(IMessage)
class Message:
    """A single status message.

    Let's make sure that this implementation actually fulfills the
    'IMessage' API.

      >>> from zope.interface.verify import verifyClass
      >>> verifyClass(IMessage, Message)
      True

      >>> status = Message(u'this is a test', type=u'info')
      >>> status.message == 'this is a test'
      True

      >>> status.type == 'info'
      True

    It is quite common to use MessageID's as status messages:

      >>> from zope.i18nmessageid import MessageFactory
      >>> from zope.i18nmessageid import Message as I18NMessage
      >>> msg_factory = MessageFactory('test')

      >>> msg = msg_factory(u'test_message', default=u'Default text')

      >>> status = Message(msg, type=u'warn')
      >>> status.type == 'warn'
      True

      >>> type(status.message) is I18NMessage
      True

      >>> status.message.default == 'Default text'
      True

      >>> status.message.domain == u'test'
      True

    """

    def __init__(self, message, type=""):
        self.message = message
        self.type = type

    def __eq__(self, other):
        if not isinstance(other, Message):
            return False
        if self.message == other.message and self.type == other.type:
            return True
        return False

    def encode(self):
        """
        Encode to a cookie friendly format.

        The format consists of a two bytes length header of 11 bits for the
        message length and 5 bits for the type length followed by two values.
        """

        fmt_tpl = "!H{0}s{1}s"
        message = _utf8(self.message)[:0x3FF]  # we can store 2^11 bytes
        type_ = _utf8(self.type)[:0x1F]  # we can store 2^5 bytes
        size = (len(message) << 5) + (len(type_) & 31)  # pack into 16 bits
        fmt = fmt_tpl.format(len(message), len(type_))
        return struct.pack(fmt, size, message, type_)


def decode(value):
    """
    Decode messages from a cookie

    We return the decoded message object, and the remainder of the cookie
    value as bytes (it can contain further messages).

    We expect at least 2 bytes (size information).
    """
    if len(value) >= 2:
        size = struct.unpack(b"!H", value[:2])[0]
        msize, tsize = (size >> 5, size & 31)
        message = Message(
            _unicode(value[2 : msize + 2]),
            _unicode(value[msize + 2 : msize + tsize + 2]),
        )
        return message, value[msize + tsize + 2 :]
    return None, b""



from pathlib import Path
from setuptools import find_packages
from setuptools import setup


version = "5.0.7"

long_description = (
    f"{Path('README.rst').read_text()}\n{Path('CHANGES.rst').read_text()}"
)

setup(
    name="Products.statusmessages",
    version=version,
    description="statusmessages provides an easy way of handling "
    "internationalized status messages managed via an "
    "BrowserRequest adapter storing status messages in "
    "client-side cookies.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    # Get more strings from
    # https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Framework :: Plone :: 6.0",
        "Framework :: Plone :: Core",
        "Framework :: Zope :: 5",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    keywords="Zope CMF Plone status messages i18n",
    author="Hanno Schlichting",
    author_email="plone-developers@lists.sourceforge.net",
    url="https://pypi.org/project/Products.statusmessages",
    license="BSD",
    packages=find_packages("src"),
    namespace_packages=["Products"],
    package_dir={"": "src"},
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.8",
    extras_require={
        "test": [],
    },
    install_requires=[
        "setuptools",
        "Zope",  # workaround for https://github.com/reinout/z3c.dependencychecker/issues/144
    ],
)
